
from binascii import hexlify


class WarnDict(dict):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            msg = self.__msg % (item,)
            self.__log.warn(msg)
            raise KeyError(msg)


class WarnList(list):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except IndexError:
            msg = self.__msg % item
            self.__log.warn(msg)
            raise IndexError(msg)


def tohex(data):
    return hexlify(data).decode('ascii')


def assert_attr(instance, *attrs):
    for attr in attrs:
        if getattr(instance, attr) is None:
            raise Exception('No %s defined' % attr)


class AttrDict(dict):

    def __getattr__(self, name):
        if name.startswith('_'):
            return super().__getattr__(name)
        else:
            try:
                return self[name]
            except KeyError:
                raise AttributeError(name)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self[name] = value



import datetime as dt
from abc import abstractmethod

from ch2.lib.date import local_date_to_time
from ch2.squeal import ActivityGroup, ActivityJournal

from .. import BasePipeline
from ...lib.date import to_date
from ...squeal.tables.pipeline import Pipeline


def build_pipeline(log, session, type, factory, date, schedule=None):
    '''
    schedule only sent for summary views.
    '''
    date = to_date(date)   # why is this needed?
    for cls, args, kargs in Pipeline.all(log, session, type):
        log.info('Building %s (%s, %s)' % (cls, args, kargs))
        yield from cls(log, *args, **kargs).build(session, factory, date, schedule=schedule)


class Displayer(BasePipeline):

    def build(self, s, f, date, schedule=None):
        if schedule:
            yield from self._build_schedule(s, f, date, schedule=schedule)
        else:
            yield from self._build_date(s, f, date)

    @abstractmethod
    def _build_schedule(self, s, f, date, schedule=None):
        raise NotImplementedError()

    @abstractmethod
    def _build_date(self, s, f, date):
        raise NotImplementedError()


class JournalDiary(Displayer):

    def _build_date(self, s, f, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for journal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield from self._journal_date(s, journal, date)

    @abstractmethod
    def _journal_date(self, s, ajournal, date):
        raise NotImplementedError()




from . import Calculator
from ...command.args import FORCE, mm
from ...squeal.tables.statistic import StatisticJournal, StatisticName


class CleanUnusedStatistics(Calculator):

    def run(self, force=False, after=None):
        if not force:
            self._log.info('Unused statistics cleaned only with %s' % mm(FORCE))
            return
        n = 0
        with self._db.session_context() as s:
            for statistic in s.query(StatisticName).outerjoin(StatisticJournal). \
                    filter(StatisticJournal.statistic_name_id == None).all():
                self._log.debug('Deleting unused %s' % statistic)
                s.delete(statistic)
                n += 1
        if n:
            self._log.warn('Deleted %d unused Statistics' % n)


# todo - empty intervals?




from ..command.args import PATH, FORCE, FAST
from ..squeal.tables.pipeline import PipelineType
from ..stoats.calculate import run_pipeline_after, run_pipeline_paths


def activities(args, log, db):
    '''
## activities

    > ch2 activities PATH [PATH ...]

Read activities data from FIT files.

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    force, fast, paths = args[FORCE], args[FAST], args[PATH]
    run_pipeline_paths(log, db, PipelineType.ACTIVITY, paths, force=force)
    if not fast:
        run_pipeline_after(log, db, PipelineType.STATISTIC, force=force)




from ..profile.profile import load_fit
from .tokens import State, FileHeader, token_factory, Checksum


def tokens(log, data, types, messages):
    state = State(log, types, messages)
    file_header = FileHeader(data)
    yield 0, file_header
    offset = len(file_header)
    file_header.validate(data)
    while len(data) - offset > 2:
        token = token_factory(data[offset:], state)
        yield offset, token
        offset += len(token)
    checksum = Checksum(data)
    yield offset, checksum
    checksum.validate(offset)


def filtered_tokens(log, fit_path, after=0, limit=-1, warn=False, profile_path=None):
    data, types, messages = load_fit(log, fit_path, warn=warn, profile_path=profile_path)

    def generator():
        for i, (offset, token) in enumerate(tokens(log, data, types, messages)):
            if i >= after and (limit < 0 or i - after < limit):
                yield i, offset, token

    return data, types, messages, generator()


def filtered_records(log, fit_path, after=0, limit=-1, records=None, warn=False, profile_path=None):
    data, types, messages = load_fit(log, fit_path, warn=warn, profile_path=profile_path)

    def generator():
        for i, (offset, token) in enumerate((offset, token)
                                            for (offset, token) in tokens(log, data, types, messages)
                                            if token.is_user):
            if i >= after and (limit < 0 or i - after < limit):
                record = token.parse(warn=warn)
                if not records or record.name in records:
                    yield record

    return data, types, messages, generator()




from .fields import Row, MessageField, TypedField
from .support import Named
from .support import Rows
from ..format.records import LazyRecord
from ...lib.data import WarnDict

HEADER_GLOBAL_TYPE = -1

HEADER_FIELDS = [
    ('header_size', 1, 'uint8'),
    ('protocol_version', 1, 'uint8'),
    ('profile_version', 1, 'uint16'),
    ('data_size', 1, 'uint32'),
    ('fit_text', 4, 'string'),
    ('checksum', 1, 'uint16')
]


class Message(Named):

    def __init__(self, log, name, number=None, warn=False):
        super().__init__(log, name)
        self.number = number
        self._profile_to_field = WarnDict(log, 'No field for profile %r') if warn else dict()
        self._number_to_field = WarnDict(log, 'No field for number %r') if warn else dict()

    def _add_field(self, field):
        self._profile_to_field[field.name] = field
        self._number_to_field[field.number] = field

    def profile_to_field(self, name):
        return self._profile_to_field[name]

    def number_to_field(self, value):
        return self._number_to_field[value]

    def _post(self, types):
        for field in self._number_to_field.values():
            field.post(self, types)

    def parse(self, data, defn, timestamp, **options):
        return LazyRecord(self.name, self.number, defn.identity, timestamp,
                          self.__parse(data, defn, timestamp, **options))

    def __parse(self, data, defn, timestamp, **options):
        # this is the generator that lives inside a record and is evaluated on demand
        references = {}
        for field in defn.fields:
            bytes = data[field.start:field.finish]
            if field.field:
                for name, value in self._parse_field(
                        field.field, bytes, field.count, defn.endian, timestamp, references,
                        defn.accumulate, self, **options):
                    if name in defn.references and value[0] is not None:
                        references[name] = value
                    yield name, value
            else:
                name = '@%d:%d' % (field.start, field.finish)
                value = (field.base_type.parse(bytes, field.count, defn.endian, timestamp), None)
                yield name, value

    def _parse_field(self, field, bytes, count, endian, timestamp, references, accumulate, message, **options):
        # allow interception for optional field in header
        yield from field.parse(bytes, count, endian, timestamp, references, accumulate, message, **options)


class RowMessage(Message):

    def __init__(self, log, row, rows, types, warn=False):
        number = types.profile_to_type('mesg_num').profile_to_internal(row.msg_name)
        super().__init__(log, row.msg_name, number, warn=warn)
        while rows:
            if not rows.peek().field_name:
                self._post(types)
                break
            self._add_field(MessageField(self._log, next(rows), rows, types))


class Header(Message):

    def __init__(self, log, types, warn=False):
        super().__init__(log, 'HEADER', number=HEADER_GLOBAL_TYPE, warn=warn)
        for n, (name, size, base_type) in enumerate(HEADER_FIELDS):
            self._add_field(TypedField(log, name, n, None, None, None, None, base_type, types))

    def _parse_field(self, field, data, count, endian, timestamp, references, accumulate, message):
        if field.name == 'checksum' and references['header_size'] == 12:
            yield None, None
        else:
            yield from super()._parse_field(field, data, count, endian, timestamp, references, accumulate, message)


class Missing(Message):

    def __init__(self, log, number):
        super().__init__(log, 'MESSAGE %d' % number, number)


class Messages:

    def __init__(self, log, sheet, types, warn=False):
        self.__log = log
        self.__profile_to_message = WarnDict(log, 'No message for profile %r') if warn else dict()
        self.__number_to_message = WarnDict(log, 'No message for number %r') if warn else dict()
        rows = Rows(sheet, wrapper=Row)
        for row in rows:
            if row.msg_name and row.msg_name[0].isupper():
                self.__log.debug('Skipping %s' % (row,))
            elif row.msg_name:
                # self.__log.info('Parsing message %s' % row.msg_name)
                self.__add_message(RowMessage(self.__log, row, rows, types, warn=warn))
        self.__add_message(Header(self.__log, types, warn=warn))

    def __add_message(self, message):
        self.__profile_to_message[message.name] = message
        try:
            self.__number_to_message[message.number] = message
        except AttributeError:
            pass

    def profile_to_message(self, name):
        return self.__profile_to_message[name]

    def number_to_message(self, number):
        try:
            return self.__number_to_message[number]
        except KeyError:
            message = Missing(self.__log, number)
            self.__number_to_message[number] = message
            return message






from ..lib.data import AttrDict
from ..squeal.types import short_cls


class BasePipeline:

    def __init__(self, log, *args, **kargs):
        self._log = log
        self.__read = set()
        self._on_init(*args, **kargs)

    def _on_init(self, *args, **kargs):
        self._args = args
        self._kargs = AttrDict(kargs)

    def _assert_karg(self, name, default=None):
        if name not in self._kargs:
            if default is None:
                raise Exception('Missing %s parameter for %s' % (name, short_cls(self)))
            else:
                self._log.warn('Using default for %s=%s' % (name, default))
                self._kargs[name] = default
                self.__read.add(name)  # avoid double logging
        value = self._kargs[name]
        if name not in self.__read:
            self._log.info('%s=%s' % (name, value))
            self.__read.add(name)
        return value


class DbPipeline(BasePipeline):

    def __init__(self, log, db, *args, **kargs):
        self._db = db
        super().__init__(log, *args, **kargs)




import datetime as dt
from json import dumps

from pendulum.tz import get_local_timezone
from sqlalchemy import Column, Integer, Text, ForeignKey
from sqlalchemy.orm import relationship, backref

from .constant import SystemConstant
from .source import SourceType, Source, Interval
from .statistic import StatisticJournal, STATISTIC_JOURNAL_CLASSES
from ..support import Base
from ..types import Date, Cls, Json, Sched, Sort
from ...lib.data import assert_attr
from ...lib.date import local_date_to_time
from ...lib.schedule import Schedule


class Topic(Base):

    __tablename__ = 'topic'

    id = Column(Integer, primary_key=True)
    parent_id = Column(Integer, ForeignKey('topic.id'), nullable=True)
    # http://docs.sqlalchemy.org/en/latest/orm/self_referential.html
    children = relationship('Topic', backref=backref('parent', remote_side=[id]))
    schedule = Column(Sched, nullable=False)
    start = Column(Date)
    finish = Column(Date)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='')

    def __init__(self, id=None, parent=None, parent_id=None, schedule=None, name=None, description=None, sort=None):
        # Topic instances are only created in config.  so we intercept here to
        # duplicate data for start and finish - it's not needed elsewhere.
        if not isinstance(schedule, Schedule):
            schedule = Schedule(schedule)
        self.id = id
        self.parent = parent
        self.parent_id = parent_id
        self.schedule = schedule
        self.name = name
        self.description = description
        self.sort = sort
        self.start = schedule.start
        self.finish = schedule.finish

    def __str__(self):
        return 'Topic "%s" (%s)' % (self.name, self.schedule)


class TopicField(Base):

    __tablename__ = 'topic_field'

    id = Column(Integer, primary_key=True)
    topic_id = Column(Integer, ForeignKey('topic.id', ondelete='cascade'), nullable=False)
    topic = relationship('Topic',
                         backref=backref('fields', cascade='all, delete-orphan',
                                         passive_deletes=True,
                                         order_by='TopicField.sort'))
    type = Column(Integer, nullable=False)  # StatisticJournalType
    sort = Column(Sort)
    statistic_name_id = Column(Integer, ForeignKey('statistic_name.id', ondelete='cascade'), nullable=False)
    statistic_name = relationship('StatisticName')
    display_cls = Column(Cls, nullable=False)
    display_args = Column(Json, nullable=False, server_default=dumps(()))
    display_kargs = Column(Json, nullable=False, server_default=dumps({}))
    schedule = Column(Sched, nullable=False, server_default='')

    def __str__(self):
        return 'TopicField "%s"/"%s"' % (self.topic.name, self.statistic_name.name)


class TopicJournal(Source):

    __tablename__ = 'topic_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    topic_id = Column(Integer, ForeignKey('topic.id'))
    topic = relationship('Topic')
    date = Column(Date, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.TOPIC
    }

    def populate(self, log, s):
        if hasattr(self, 'statistics'):
            return
        assert_attr(self, 'date')
        if self.id is None:
            s.flush()
        log.debug('Populating journal for topic %s at %s' % (self.topic.name, self.date))
        self.statistics = {}
        for field in self.topic.fields:
            assert_attr(field, 'schedule')
            if field.schedule.at_location(self.date):
                log.debug('Finding StatisticJournal for field %s' % field.statistic_name.name)
                journal = StatisticJournal.at_date(s, self.date, field.statistic_name.name,
                                                   field.statistic_name.owner, self.topic)
                if not journal:
                    journal = STATISTIC_JOURNAL_CLASSES[field.type](
                        statistic_name=field.statistic_name, source=self, time=local_date_to_time(self.date))
                    s.add(journal)
                self.statistics[field] = journal

    def __str__(self):
        return 'TopicJournal from %s' % self.date

    @classmethod
    def check_tz(cls, log, s):
        from ..database import add
        tz = get_local_timezone()
        db_tz = s.query(SystemConstant).filter(SystemConstant.name == SystemConstant.TIMEZONE).one_or_none()
        if not db_tz:
            db_tz = add(s, SystemConstant(name=SystemConstant.TIMEZONE, value=''))
        if db_tz.value != tz.name:
            cls.__reset_timezone(log, s)
            db_tz.value = tz.name

    @classmethod
    def __reset_timezone(cls, log, s):
        log.warn('Timezone has changed')
        log.warn('Recalculating times for TopicJournal entries')
        for tj in s.query(TopicJournal).all():
            tj.time = local_date_to_time(tj.date)
        Interval.delete_all(log, s)

    def time_range(self, s):
        start = local_date_to_time(self.date)
        return start, start + dt.timedelta(days=1)




from sqlalchemy import Column, Text, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..support import Base
from ..types import Time, Sort
from ...lib.date import format_time


class ActivityGroup(Base):

    __tablename__ = 'activity_group'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='')

    def __str__(self):
        return 'ActivityGroup "%s"' % self.name


class ActivityJournal(Source):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    activity_group_id = Column(Integer, ForeignKey('activity_group.id'), nullable=False)
    activity_group = relationship('ActivityGroup')
    name = Column(Text, unique=True)
    fit_file = Column(Text, nullable=False, unique=True)
    start = Column(Time, nullable=False, index=True)
    finish = Column(Time, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'ActivityJournal %s %s to %s' % (self.activity_group.name,
                                                format_time(self.start), format_time(self.finish))

    def time_range(self, s):
        return self.start, self.finish


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False, index=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint(activity_journal_id, start)

    def __str__(self):
        return 'ActivityTimespan from %s - %s' % (format_time(self.start), format_time(self.finish))






from collections import namedtuple
from json import loads

from sqlalchemy import desc, inspect, select, and_

from . import ActivityCalculator
from ..load import StatisticJournalLoader
from ..names import FTHR, HR_ZONE, HEART_RATE, S
from ...squeal.tables.activity import ActivityGroup
from ...squeal.tables.constant import Constant
from ...squeal.tables.statistic import StatisticJournal, StatisticName, StatisticJournalFloat, StatisticJournalInteger

# constraint comes from constant
HRImpulse = namedtuple('HRImpulse', 'dest_name, gamma, zero, max_secs')

# values from british cycling online calculator
# these are upper limits
BC_ZONES = (68, 83, 94, 105, 121, 999)


def hr_zones_from_database(log, s, activity_group, time):
    fthr = StatisticJournal.before(s, time, FTHR, Constant, activity_group)
    if fthr:
        return hr_zones(fthr.value)
    else:
        return None


def hr_zones(fthr):
    return [fthr * pc / 100.0 for pc in BC_ZONES]


class HeartRateStatistics(ActivityCalculator):

    def __init__(self, log, db, *args, **kargs):
        self.__fthr_cache = None
        # must create the statistic(s) before finding out if it is missing...
        with db.session_context() as s:
            for agroup in s.query(ActivityGroup).all():
                StatisticName.add_if_missing(log, s, HR_ZONE, None, None, self, agroup)
        super().__init__(log, db, *args, **kargs)

    def _filter_statistic_journals(self, q):
        return q.filter(StatisticName.name == HR_ZONE)

    def _add_stats(self, s, ajournal):

        impulse = self._assert_karg('impulse')
        hr_impulse = HRImpulse(**loads(Constant.get(s, impulse).at(s).value))
        self._log.debug('%s: %s' % (impulse, hr_impulse))

        loader = StatisticJournalLoader(self._log, s, self)

        sn = inspect(StatisticName).local_table
        sj = inspect(StatisticJournal).local_table
        sji = inspect(StatisticJournalInteger).local_table

        stmt = select([sj.c.time, sji.c.value]). \
            select_from(sj.join(sn).join(sji)). \
            where(and_(sj.c.source_id == ajournal.id,
                       sn.c.owner == self._assert_karg('owner'),
                       sn.c.constraint == ajournal.activity_group,
                       sn.c.name == HEART_RATE)). \
            order_by(sj.c.time)
        # self._log.debug(stmt)

        prev_time, prev_heart_rate_zone = None, None
        for time, heart_rate in s.connection().execute(stmt):
            if heart_rate:
                heart_rate_zone = self._calculate_zone(s, heart_rate, time, ajournal.activity_group)
                if heart_rate_zone is not None:
                    loader.add(HR_ZONE, None, None, ajournal.activity_group, ajournal, heart_rate_zone, time,
                               StatisticJournalFloat)
            else:
                heart_rate_zone = None
            if prev_heart_rate_zone is not None:
                duration = (time - prev_time).total_seconds()
                if duration <= hr_impulse.max_secs:
                    heart_rate_impulse = self._calculate_impulse(prev_heart_rate_zone, duration, hr_impulse)
                    loader.add(hr_impulse.dest_name, None, None, ajournal.activity_group, ajournal,
                               heart_rate_impulse, time, StatisticJournalFloat)
                    loader.add('%s (duration)' % hr_impulse.dest_name, S, None, ajournal.activity_group, ajournal,
                               duration, time, StatisticJournalFloat)
            prev_time, prev_heart_rate_zone = time, heart_rate_zone

        # if there are no values, add a single null so we don't re-process
        if not loader:
            loader.add(HR_ZONE, None, None, ajournal.activity_group, ajournal, None, ajournal.start,
                       StatisticJournalFloat)

        loader.load()
        s.commit()

    def _calculate_impulse(self, heart_rate_zone, duration, hr_impulse):
        return duration * ((max(heart_rate_zone, hr_impulse.zero) - hr_impulse.zero)
                           / (6 - hr_impulse.zero)) ** hr_impulse.gamma

    def _calculate_zone(self, s, heart_rate, time, activity_group):
        self._load_fthr_cache(s, activity_group)
        for fthr in self.__fthr_cache:
            if fthr.time <= time:
                lower_limit, prev_delta = 0, None
                zones = hr_zones(fthr.value)
                for zone, upper_limit in enumerate(zones):
                    if lower_limit <= heart_rate < upper_limit:
                        if zone == 0:
                            return 1 + zone
                        elif zone == 5:
                            return 1 + zone + (heart_rate - lower_limit) / prev_delta
                        else:
                            return 1 + zone + (heart_rate - lower_limit) / (upper_limit - lower_limit)
                    prev_delta = upper_limit - lower_limit
                    lower_limit = upper_limit

    def _load_fthr_cache(self, s, activity_group):
        if self.__fthr_cache is None:
            self.__fthr_cache = list(s.query(StatisticJournal).join(StatisticName).
                                     filter(StatisticName.name == FTHR,
                                            StatisticName.owner == Constant,
                                            StatisticName.constraint == activity_group).
                                     order_by(desc(StatisticJournal.time)).all())





from .tables.activity import ActivityGroup, ActivityTimespan, ActivityJournal
from .tables.constant import SystemConstant, Constant
from .tables.monitor import MonitorJournal
from .tables.nearby import ActivitySimilarity, ActivityNearby
from .tables.pipeline import Pipeline
from .tables.segment import Segment, SegmentJournal
from .tables.source import Source, Interval
from .tables.statistic import StatisticName, StatisticJournalFloat, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournal, StatisticMeasure
from .tables.topic import TopicJournal, Topic





from io import BytesIO
from os.path import join, exists
from re import search, sub
from zipfile import ZipFile

from requests import session


class GarminConnect:

    # logic and data from https://github.com/tcgoetz/GarminDB/blob/master/download_garmin.py
    # and also a bunch more similar scripts

    base_url = 'https://connect.garmin.com'
    sso_url = 'https://sso.garmin.com/sso'
    modern = base_url + '/modern'
    signin = sso_url + '/signin'
    daily = modern + '/proxy/download-service/files/wellness'

    headers = {
        # 'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:62.0) Gecko/20100101 Firefox/62.0'
        'User-Agent': 'Mozilla/5.0 (compatible; https://github.com/andrewcooke/choochoo)'
    }

    def __init__(self, log, log_response=False):
        self._log = log
        self._r = session()
        self._log_response = log_response

    def login(self, username, password):

        self._log.info('Connecting to Garmin Connect as %s' % username)

        params = {
            # todo - are these all actually needed?  by an sso service?
            'webhost': self.base_url,
            'service': self.modern,
            'source': self.signin,
            'redirectAfterAccountLoginUrl': self.modern,
            'redirectAfterAccountCreationUrl': self.modern,
            'gauthHost': self.sso_url,
            'locale': 'en_US',
            'id': 'gauth-widget',
            'cssUrl': 'https://static.garmincdn.com/com.garmin.connect/ui/css/gauth-custom-v1.2-min.css',
            'clientId': 'GarminConnect',
            'rememberMeShown': 'true',
            'rememberMeChecked': 'false',
            'createAccountShown': 'true',
            'openCreateAccount': 'false',
            'usernameShown': 'false',
            'displayNameShown': 'false',
            'consumeServiceTicket': 'false',
            'initialFocus': 'true',
            'embedWidget': 'false',
            'generateExtraServiceTicket': 'false'
        }

        response = self._log_r(self._r.get(self.signin, headers=self.headers, params=params))
        response.raise_for_status()

        data = {
            'username': username,
            'password': password,
            'embed': 'true',
            'lt': 'e1s1',
            '_eventId': 'submit',
            'displayNameRequired': 'false'
        }

        response = self._log_r(self._r.post(self.signin, headers=self.headers, params=params, data=data))
        response.raise_for_status()

        response_url = search(r'"(https:[^"]+?ticket=[^"]+)"', response.text)
        if not response_url:
            raise Exception('Could not find response URL')
        response_url = sub(r'\\', '', response_url.group(1))
        self._log.debug('Response URL: %s' % response_url)

        response = self._log_r(self._r.get(response_url))
        response.raise_for_status()

    def get_monitoring(self, date):
        response = self._log_r(self._r.get(self.daily + date.strftime('/%Y-%m-%d'), headers=self.headers))
        response.raise_for_status()
        return response

    def _log_r(self, response):
        if self._log_response:
            self._log.debug('headers: %s' % response.headers)
            self._log.debug('reason: %s' % response.reason)
            self._log.debug('cookies: %s' % response.cookies)
            self._log.debug('history: %s' % response.history)
            # self._log.debug('text: %s' % response.text)
        return response

    def get_monitoring_to_zip_file(self, date, dir):
        path = join(dir, date.strftime('%Y-%m-%d.zip'))
        if exists(path):
            raise Exception('"%s" already exists' % path)
        response = self.get_monitoring(date)
        with open(path, 'wb') as f:
            f.write(response.content)
        self._log.info('Downloaded data for %s to %s' % (date, path))

    def get_monitoring_to_fit_file(self, date, dir):
        response = self.get_monitoring(date)
        zipfile = ZipFile(BytesIO(response.content))
        for name in zipfile.namelist():
            path = zipfile.extract(name, path=dir)
            self._log.info('Downloaded data for %s to %s' % (date, path))




from collections import defaultdict, namedtuple
from json import loads

from sqlalchemy import inspect, select, alias, and_, distinct, func

from .. import DbPipeline
from ..names import LONGITUDE, LATITUDE
from ...arty import MatchType
from ...arty.spherical import SQRTree
from ...lib.date import to_time
from ...lib.dbscan import DBSCAN
from ...lib.optimizn import expand_max
from ...squeal.tables.activity import ActivityJournal, ActivityGroup
from ...squeal.tables.constant import Constant
from ...squeal.tables.nearby import ActivitySimilarity, ActivityNearby
from ...squeal.tables.statistic import StatisticName, StatisticJournal, StatisticJournalFloat

Nearby = namedtuple('Nearby', 'constraint, activity_group, border, start, finish, '
                              'latitude, longitude, height, width')


class NearbySimilarityCalculator(DbPipeline):

    def _on_init(self, *args, **kargs):
        super()._on_init(*args, **kargs)
        nearby = self._assert_karg('nearby')
        with self._db.session_context() as s:
            self._config = Nearby(**loads(Constant.get(s, nearby).at(s).value))
        self._log.info('%s: %s' % (nearby, self._config))

    def run(self, force=False, after=None):

        rtree = SQRTree(default_match=MatchType.INTERSECTS, default_border=self._config.border)

        with self._db.session_context() as s:
            if force:
                self._delete(s)
            n_points = defaultdict(lambda: 0)
            self._prepare(s, rtree, n_points, 10000)
            n_intersects = defaultdict(lambda: defaultdict(lambda: 0))
            new_ids = self._measure(s, rtree, n_points, n_intersects, 1000)
            self._save(s, new_ids, n_points, n_intersects, 10000)

    def _delete(self, s):
        self._log.warn('Deleting similarity data for %s' % self._config.constraint)
        s.query(ActivitySimilarity). \
            filter(ActivitySimilarity.constraint == self._config.constraint). \
            delete()

    def _save(self, s, new_ids, n_points, n_intersects, delta):
        n, total = 0, (len(new_ids) * (len(new_ids) - 1)) / 2
        for lo in new_ids:
            for hi in filter(lambda hi: hi > lo, new_ids):
                s.add(ActivitySimilarity(constraint=self._config.constraint,
                                         activity_journal_lo_id=lo, activity_journal_hi_id=hi,
                                         similarity=n_intersects[lo][hi] / (n_points[lo] + n_points[hi])))
                n += 1
                if n % delta == 0:
                    self._log.info('Saved %d / %d for %s' % (n, total, self._config.constraint))

    def _prepare(self, s, rtree, n_points, delta):
        n = 0
        for id_in, lon, lat in self._data(s, new=False):
            p = [(lon, lat)]
            rtree[p] = id_in
            n_points[id_in] += 1
            n += 1
            if n % delta == 0:
                self._log.info('Loaded %s points for %s' % (n, self._config.constraint))

    def _measure(self, s, rtree, n_points, n_intersects, delta):
        new_ids, current_id, seen, n = [], None, None, 0
        for id_in, lon, lat in self._data(s, new=True):
            if id_in != current_id:
                current_id, seen = id_in, set()
                new_ids.append(id_in)
            p = [(lon, lat)]
            for other_p, id_out in rtree.get_items(p):
                if id_in != id_out:
                    if other_p not in seen:
                        lo, hi = min(id_in, id_out), max(id_in, id_out)
                        n_intersects[lo][hi] += 1
                        seen.add(other_p)
            rtree[p] = id_in
            n_points[id_in] += 1
            n += 1
            if n % delta == 0:
                self._log.info('Measured %s points for %s' % (n, self._config.constraint))
        return new_ids

    def _data(self, s, new=True):

        start = to_time(self._config.start)
        finish = to_time(self._config.finish)

        lat = s.query(StatisticName.id).filter(StatisticName.name == LATITUDE).scalar()
        lon = s.query(StatisticName.id).filter(StatisticName.name == LONGITUDE).scalar()
        agroup = s.query(ActivityGroup.id).filter(ActivityGroup.name == self._config.activity_group).scalar()

        sj_lat = inspect(StatisticJournal).local_table
        sj_lon = alias(inspect(StatisticJournal).local_table)
        sjf_lat = inspect(StatisticJournalFloat).local_table
        sjf_lon = alias(inspect(StatisticJournalFloat).local_table)
        aj = inspect(ActivityJournal).local_table
        ns = inspect(ActivitySimilarity).local_table

        existing_lo = select([ns.c.activity_journal_lo_id]). \
            where(ns.c.constraint == self._config.constraint)
        existing_hi = select([ns.c.activity_journal_hi_id]). \
            where(ns.c.constraint == self._config.constraint)
        existing = existing_lo.union(existing_hi).cte()

        stmt = select([sj_lat.c.source_id, sjf_lon.c.value, sjf_lat.c.value]). \
            select_from(sj_lat).select_from(sj_lon).select_from(sjf_lat).select_from(sjf_lat).select_from(aj). \
            where(and_(sj_lat.c.source_id == sj_lon.c.source_id,  # same source
                       sj_lat.c.time == sj_lon.c.time,            # same time
                       sj_lat.c.source_id == aj.c.id,             # and associated with an activity
                       aj.c.activity_group_id == agroup,          # of the right group
                       sj_lat.c.id == sjf_lat.c.id,               # lat sub-class
                       sj_lon.c.id == sjf_lon.c.id,               # lon sub-class
                       sj_lat.c.statistic_name_id == lat,         # lat name
                       sj_lon.c.statistic_name_id == lon,         # lon name
                       sj_lat.c.time >= start.timestamp(),        # time limits
                       sj_lat.c.time < finish.timestamp(),
                       sjf_lat.c.value > self._config.latitude - self._config.height / 2,
                       sjf_lat.c.value < self._config.latitude + self._config.height / 2,
                       sjf_lon.c.value > self._config.longitude - self._config.width / 2,
                       sjf_lon.c.value < self._config.longitude + self._config.width / 2))

        if new:
            stmt = stmt.where(func.not_(sj_lat.c.source_id.in_(existing)))
        else:
            stmt = stmt.where(sj_lat.c.source_id.in_(existing))
        stmt = stmt.order_by(sj_lat.c.source_id)  # needed for seen logic
        yield from s.connection().execute(stmt)


class NearbySimilarityDBSCAN(DBSCAN):

    def __init__(self, log, s, constraint, epsilon, minpts):
        super().__init__(log, epsilon, minpts)
        self.__s = s
        self.__constraint = constraint
        self.__max_similarity = self.__s.query(func.max(ActivitySimilarity.similarity)). \
            filter(ActivitySimilarity.constraint == constraint).scalar()
        # self._log.info('Max similarity %.2f' % self.__max_similarity)

    def run(self):
        candidates = set(x[0] for x in
                         self.__s.query(distinct(ActivitySimilarity.activity_journal_lo_id)).
                         filter(ActivitySimilarity.constraint == self.__constraint).all())
        candidates.union(set(x[0] for x in
                             self.__s.query(distinct(ActivitySimilarity.activity_journal_lo_id)).
                             filter(ActivitySimilarity.constraint == self.__constraint).all()))
        candidates = sorted(candidates)
        # shuffle(candidates)  # skip for repeatability
        return super().run(candidates)

    def neighbourhood(self, candidate, epsilon):
        qlo = self.__s.query(ActivitySimilarity.activity_journal_lo_id). \
            filter(ActivitySimilarity.constraint == self.__constraint,
                   ActivitySimilarity.activity_journal_hi_id == candidate,
                   (self.__max_similarity - ActivitySimilarity.similarity) / self.__max_similarity < epsilon)
        qhi = self.__s.query(ActivitySimilarity.activity_journal_hi_id). \
            filter(ActivitySimilarity.constraint == self.__constraint,
                   ActivitySimilarity.activity_journal_lo_id == candidate,
                   (self.__max_similarity - ActivitySimilarity.similarity) / self.__max_similarity < epsilon)
        return [x[0] for x in qlo.all()] + [x[0] for x in qhi.all()]


class NearbyStatistics(NearbySimilarityCalculator):

    def run(self, force=False, after=None):
        super().run(force=force, after=after)
        with self._db.session_context() as s:
            d_min, _ = expand_max(self._log, 0, 1, 5, lambda d: len(self.dbscan(s, d)))
            self.save(s, self.dbscan(s, d_min))

    def dbscan(self, s, d):
        return NearbySimilarityDBSCAN(self._log, s, self._config.constraint, d, 3).run()

    def save(self, s, groups):
        s.query(ActivityNearby). \
            filter(ActivityNearby.constraint == self._config.constraint).delete()
        for i, group in enumerate(groups):
            self._log.info('Group %d has %d members' % (i, len(group)))
            for activity_journal_id in group:
                s.add(ActivityNearby(constraint=self._config.constraint, group=i,
                                     activity_journal_id=activity_journal_id))




import datetime as dt
from abc import abstractmethod

from sqlalchemy import or_
from urwid import MainLoop, Columns, Pile, Frame, Filler, Text, Divider, WEIGHT

from .args import DATE, SCHEDULE
from ..lib.date import to_date
from ..lib.io import tui
from ..lib.schedule import Schedule
from ..lib.utils import PALETTE_RAINBOW, em
from ..lib.widgets import DateSwitcher
from ..squeal.database import add
from ..squeal.tables.pipeline import PipelineType
from ..squeal.tables.topic import Topic, TopicJournal
from ..stoats.display import build_pipeline
from ..uweird.fields import PAGE_WIDTH
from ..uweird.fields.summary import summary_columns
from ..uweird.tui.decorators import Border, Indent
from ..uweird.tui.factory import Factory
from ..uweird.tui.tabs import TabList
from ..uweird.tui.widgets import DividedPile


@tui
def diary(args, log, db):
    '''
## diary

    > ch2 diary [DATE]

The date can be an absolute day or the number of days previous.  So `ch2 diary 1` selects yesterday.

Display the daily diary.  Enter information here.

To exit, alt-q (or, without saving, alt-x).

    > ch2 diary (--month | --year | --schedule SCHEDULE) [DATE}

Display a summary for the month / year / schedule.
    '''
    date, schedule = args[DATE], args[SCHEDULE]
    if not date:
        date = dt.date.today()
    else:
        try:
            date = to_date(date)
        except:
            date = dt.date.today() - dt.timedelta(days=int(date))
    with db.session_context() as s:
        TopicJournal.check_tz(log, s)
    if schedule:
        schedule = Schedule(schedule)
        if schedule.start or schedule.finish:
            raise Exception('Schedule must be open (no start or finish)')
        MainLoop(ScheduleDiary(log, db, date, schedule), palette=PALETTE_RAINBOW).run()
    else:
        MainLoop(DailyDiary(log, db, date), palette=PALETTE_RAINBOW).run()


class Diary(DateSwitcher):

    def _build(self, s):
        self._log.debug('Building diary at %s' % self._date)
        body, f = [], Factory(TabList())
        root_topics = list(self._topics(s))
        for topic in root_topics:
            body.append(self.__display_topic(s, f, topic))
        for extra in self._display_pipeline(s, f):
            body.append(extra)
        body = Border(Frame(Filler(DividedPile(body), valign='top'),
                            header=Pile([self._header(), Divider()]),
                            footer=Pile([Divider(), Text(self.__footer(), align='center')])))
        return body, f.tabs

    @abstractmethod
    def _header(self):
        pass

    def __footer(self):
        footer = ['meta-', em('q'), 'uit/e', em('x'), 'it/', em('s'), 'ave']
        footer += [' [shift]meta-']
        for sep, c in enumerate('dwmya'):
            if sep: footer += ['/']
            footer += [em(c)]
        footer += ['ctivity/', em('t'), 'oday']
        return footer

    def __display_topic(self, s, f, topic):
        self._log.debug('%s' % topic)
        body, title = [], None
        if topic.name:
            title = Text(topic.name)
        if topic.description:
            body.append(Text(topic.description))
        body += list(self._display_fields(s, f, topic))
        body += list(self.__display_children(s, f, topic))
        if not body:
            return title
        body = Indent(Pile(body))
        if title:
            body = Pile([title, body])
        return body

    def __display_children(self, s, f, topic):
        for child in topic.children:
            if self._filter_child(child):
                extra = self.__display_topic(s, f, child)
                if extra:
                    yield extra

    def _filter_child(self, child):
        return child.schedule.at_location(self._date)


class DailyDiary(Diary):
    '''
    Render the diary at a given date.
    '''

    def _topics(self, s):
        for topic in s.query(Topic).filter(Topic.parent == None,
                                           or_(Topic.start <= self._date, Topic.start == None),
                                           or_(Topic.finish >= self._date, Topic.finish == None)). \
                order_by(Topic.sort).all():
            if topic.schedule.at_location(self._date):
                yield topic

    def _header(self):
        return Text(self._date.strftime('%Y-%m-%d - %A'))

    def _display_fields(self, s, f, topic):
        columns, width = [], 0
        tjournal = self.__topic_journal(s, topic)
        tjournal.populate(self._log, s)
        for field in topic.fields:
            if field in tjournal.statistics:  # might be outside schedule
                self._log.debug('%s' % field)
                display = field.display_cls(self._log, tjournal.statistics[field],
                                            *field.display_args, **field.display_kargs)
                self._log.debug('%s' % display)
                if width + display.width > PAGE_WIDTH:
                    yield Columns(columns)
                    columns, width = [], 0
                columns.append((WEIGHT, display.width, f(display.widget())))
                width += display.width
        if width:
            yield Columns(columns)

    def __topic_journal(self, s, topic):
        tjournal = s.query(TopicJournal). \
            filter(TopicJournal.topic == topic,
                   TopicJournal.date == self._date).one_or_none()
        if not tjournal:
            tjournal = add(s, TopicJournal(topic=topic, date=self._date))
        return tjournal

    def _display_pipeline(self, s, f):
        yield from build_pipeline(self._log, s, PipelineType.DIARY, f, self._date)


class ScheduleDiary(Diary):
    '''
    Display summary data for the given schedule.
    '''

    def __init__(self, log, db, date, schedule):
        self._schedule = schedule
        super().__init__(log, db, self._new_date(date))

    def _header(self):
        return Text(self._date.strftime('%Y-%m-%d') + ' - Summary for %s' % self._schedule.describe())

    def _new_date(self, date):
        return self._schedule.start_of_frame(date)

    def _topics(self, s):
        finish = self._schedule.next_frame(self._date)
        return s.query(Topic).filter(Topic.parent == None,
                                     or_(Topic.start < finish, Topic.start == None),
                                     or_(Topic.finish >= self._date, Topic.finish == None)). \
            order_by(Topic.sort).all()

    def _filter_child(self, child):
        finish = self._schedule.next_frame(self._date)
        return (child.start is None or child.start < finish) and (child.finish is None or child.finish > self._date)

    def _display_fields(self, s, f, topic):
        names = [field.statistic_name for field in topic.fields]
        yield from summary_columns(self._log, s, f, self._date, self._schedule, names)

    def _display_pipeline(self, s, f):
        yield from build_pipeline(self._log, s, PipelineType.DIARY, f, self._date, schedule=self._schedule)





from urwid import Pile, Text

from . import JournalDiary
from ...lib.utils import label
from ...squeal import ActivityNearby
from ...uweird.tui.decorators import Indent


class NearbyDiary(JournalDiary):

    def _build_schedule(self, s, f, date, schedule=None):
        yield from []

    def _journal_date(self, s, ajournal, date):
        for group in s.query(ActivityNearby). \
                filter(ActivityNearby.activity_journal == ajournal):
            nearby = s.query(ActivityNearby). \
                filter(ActivityNearby.constraint == group.constraint,
                       ActivityNearby.group == group.group,
                       ActivityNearby.activity_journal != ajournal).all()
            dates = self._build_dates(nearby)
            yield Pile([Text([label('Nearby: '), group.constraint]),
                        Indent(Text(dates))])

    def _build_dates(self, nearby):
        dates = [d.strftime('%Y-%m-%d')
                 for d in sorted(n.activity_journal.start for n in nearby)]
        if len(dates) < 7:
            return ' '.join(dates)
        else:
            return ' '.join(dates[:3]) + ' ... ' + ' '.join(dates[-3:])




from collections import defaultdict

import pandas as pd
from sqlalchemy import inspect, select
from sqlalchemy.sql.functions import coalesce

from ..squeal.database import connect
from ..squeal import ActivityJournal, StatisticName, StatisticJournal, StatisticJournalInteger, \
    StatisticJournalFloat, StatisticJournalText, Interval, StatisticMeasure, Source
from ..squeal.types import short_cls
from ..stoats.read.segment import SegmentImporter
from ..stoats.waypoint import WaypointReader


LOG = [None]


def df(query):
    # https://stackoverflow.com/questions/29525808/sqlalchemy-orm-conversion-to-pandas-dataframe
    return pd.read_sql(query.statement, query.session.bind)


def session(*args):
    ns, log, db = connect(args)
    LOG[0] = log
    return db.session()


def log():
    if not LOG[0]:
        raise Exception('Create session first')
    return LOG[0]


def waypoints(s, activity_journal_id, *statistics, owner=short_cls(SegmentImporter)):
    names = dict((statistic, statistic) for statistic in statistics)
    # int() to convert numpy types
    ajournal = s.query(ActivityJournal).filter(ActivityJournal.id == int(activity_journal_id)).one()
    return pd.DataFrame(WaypointReader(log(), with_timespan=False).read(s, ajournal, names, owner))


def _collect_statistics(s, names):
    if not names:
        names = ['%']
    statistic_ids, statistic_names = set(), set()
    for name in names:
        for statistic_name in s.query(StatisticName). \
                filter(StatisticName.name.like(name)).all():
            statistic_ids.add(statistic_name.id)
            statistic_names.add(statistic_name.name)
    return statistic_names, statistic_ids


def _build_statistic_journal_query(statistic_ids, start, finish, owner, constraint, source_ids, schedule):

    # use more efficient expression interface and exploit the fact that
    # alternative journal types will be null in an outer join.

    sj = inspect(StatisticJournal).local_table
    sn = inspect(StatisticName).local_table
    sji = inspect(StatisticJournalInteger).local_table
    sjf = inspect(StatisticJournalFloat).local_table
    sjt = inspect(StatisticJournalText).local_table
    inv = inspect(Interval).local_table

    q = select([sn.c.name, sj.c.time, coalesce(sjf.c.value, sji.c.value, sjt.c.value)]). \
        select_from(sj.join(sn).outerjoin(sjf).outerjoin(sji).outerjoin(sjt)). \
        where(sn.c.id.in_(statistic_ids))
    if start:
        q = q.where(sj.c.time >= start)
    if finish:
        q = q.where(sj.c.time <= finish)
    if owner:
        q = q.where(sn.c.owner == owner)
    if constraint:
        q = q.where(sn.c.constraint == constraint)
    if source_ids is not None:  # avoid testing DataFrame sequences as bools
        # int() to convert numpy types
        q = q.where(sj.c.source_id.in_(int(id) for id in source_ids))
    if schedule:
        q = q.join(inv).where(inv.c.schedule == schedule)
    q = q.order_by(sj.c.time)
    return q


def statistics(s, *statistics,
               start=None, finish=None, owner=None, constraint=None, source_ids=None, schedule=None):
    statistic_names, statistic_ids = _collect_statistics(s, statistics)
    q = _build_statistic_journal_query(statistic_ids, start, finish, owner, constraint, source_ids, schedule)
    data, times, err_cnt = defaultdict(list), [], defaultdict(lambda: 0)

    def pad():
        n = len(times)
        for name in statistic_names:
            if len(data[name]) != n:
                err_cnt[name] += 1
                if err_cnt[name] <= 1:
                    log().warn('Missing %s at %s (only warning for this name)' % (name, times[-1]))
                data[name].append(None)

    for name, time, value in s.connection().execute(q):
        if times and times[-1] != time:
            pad()
        if not times or times[-1] != time:
            times.append(time)
        if len(data[name]) >= len(times):
            raise Exception('Duplicate data for %s at %s ' % (name, time) +
                            '(you may need to specify more constraints to make the query unique)')
        data[name].append(value)
    pad()
    return pd.DataFrame(data, index=times)


def statistic_quartiles(s, *statistics, schedule='m',
                        start=None, finish=None, owner=None, constraint=None, source_ids=None):
    statistic_names, statistic_ids = _collect_statistics(s, statistics)
    q = s.query(StatisticMeasure). \
        join(StatisticJournal, StatisticMeasure.statistic_journal_id == StatisticJournal.id). \
        join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
        join(Source, StatisticJournal.source_id == Source.id). \
        filter(StatisticName.id.in_(statistic_ids)). \
        filter(StatisticMeasure.quartile != None)
    if start:
        q = q.filter(StatisticJournal.time >= start)
    if finish:
        q = q.filter(StatisticJournal.time <= finish)
    if owner:
        q = q.filter(StatisticName.owner == owner)
    if constraint:
        q = q.filter(StatisticName.constraint == constraint)
    if source_ids is not None:
        q = q.filter(StatisticJournal.source_id.in_(source_ids))
    if schedule:
        q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
            filter(Interval.schedule == schedule)
    log().debug(q)
    raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
    for measure in q.all():
        raw_data[measure.source.start][measure.statistic_journal.statistic_name.name][measure.quartile] = \
            measure.statistic_journal.value
    data, times = defaultdict(list), []
    for time in sorted(raw_data.keys()):
        times.append(time)
        sub_data = raw_data[time]
        for statistic in statistic_names:
            if statistic in sub_data:
                data[statistic].append(sub_data[statistic])
            else:
                data[statistic].append(None)
    return pd.DataFrame(data, index=times)





from tempfile import NamedTemporaryFile

from sqlalchemy.sql.functions import count

from ch2.command.args import bootstrap_file, m, V, DEV, mm
from ch2.command.constants import constants
from ch2.config.default import default
from ch2.squeal.tables.constant import Constant


def test_activities():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        with db.session_context() as s:

            n = s.query(count(Constant.id)).scalar()
            assert n == 5, n

        args, log, db = bootstrap_file(f, m(V), '5', 'constants', '--set', 'FTHR.%', '154')
        constants(args, log, db)

        with db.session_context() as s:

            n = s.query(count(Constant.id)).scalar()
            assert n == 5, n
            # todo - maybe test for value?





# http://acooke.org/cute/EfficientS2.html
def expand_max(log, lo, hi, n, f):
    data = [(x, f(x)) for x in (lo + i * (hi - lo) / 4 for i in range(5))]
    for _ in range(n):
        log.info('%s' % data)
        while len(data) > 3:
            w = sum(x*fx for (i, (x, fx)) in enumerate(data)) / sum(fx for (x, fx) in data)
            m = sum(x for (x, fx) in data) / len(data)
            if w > m:
                del data[0]
            else:
                del data[-1]
        x = (data[0][0] + data[1][0]) / 2
        data.insert(1, (x, f(x)))
        x = (data[-2][0] + data[-1][0]) / 2
        data.insert(-1, (x, f(x)))
    x_max, fx_max = None, None
    for (x, fx) in data:
        if x_max is None or fx > fx_max:
            x_max, fx_max = x, fx
    return x_max, fx_max




from os.path import splitext, basename

from pygeotile.point import Point

from ..load import StatisticJournalLoader
from ..names import LATITUDE, DEG, LONGITUDE, HEART_RATE, DISTANCE, KMH, SPEED, BPM, M, SPHERICAL_MERCATOR_X, \
    SPHERICAL_MERCATOR_Y
from ..read import AbortImport, Importer
from ...fit.format.read import filtered_records
from ...fit.format.records import fix_degrees
from ...lib.date import to_time
from ...squeal.database import add
from ...squeal.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...squeal.tables.source import Interval
from ...squeal.tables.statistic import StatisticJournalFloat, StatisticJournalInteger


class ActivityImporter(Importer):

    def run(self, paths, force=False):
        if 'sport_to_activity' not in self._kargs:
            raise Exception('No map from sport to activity')
        self._run(paths, force=force)

    def _activity_group(self, s, path, sport, sport_to_activity):
        if sport in sport_to_activity:
            return self._lookup_activity_group(s, sport_to_activity[sport])
        else:
            self._log.warn('Unrecognised sport: "%s" in %s' % (sport, path))
            raise AbortImport()

    def _lookup_activity_group(self, s, name):
        activity_group = s.query(ActivityGroup).filter(ActivityGroup.name == name).one_or_none()
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                self._log.info('Available activity group:')
                for activity_group in activities:
                    self._log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                self._log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _delete_journals(self, s, activity_group, first_timestamp):
        # need to iterate because sqlite doesn't support multi-table delete and we have inheritance.
        for journal in s.query(ActivityJournal). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.start == first_timestamp).all():
            s.delete(journal)
        s.flush()

    def _import(self, s, path):

        sport_to_activity = self._assert_karg('sport_to_activity')
        loader = StatisticJournalLoader(self._log, s, self)

        data, types, messages, records = filtered_records(self._log, path)
        records = [record.force(fix_degrees)
                   for record in sorted(records, key=lambda r: r.timestamp if r.timestamp else to_time(0.0))]

        first_timestamp = self._first(path, records, 'event', 'record').value.timestamp
        sport = self._first(path, records, 'sport').value.sport.lower()
        activity_group = self._activity_group(s, path, sport, sport_to_activity)
        self._delete_journals(s, activity_group, first_timestamp)
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          fit_file=path, name=splitext(basename(path))[0]))

        timespan, warned, last_timestamp = None, 0, to_time(0.0)
        self._log.info('Importing activity data from %s' % path)
        for record in records:
            if record.name == 'event' or (record.name == 'record' and record.timestamp > last_timestamp):
                if record.name == 'event' and record.value.event == 'timer' and record.value.event_type == 'start':
                    if timespan:
                        self._log.warn('Ignoring start with no corresponding stop (possible lost data?)')
                    else:
                        timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                           start=record.value.timestamp,
                                                           finish=record.value.timestamp))
                if record.name == 'record':
                    loader.add(LATITUDE, DEG, None, activity_group, ajournal,
                              record.none.position_lat, record.value.timestamp, StatisticJournalFloat)
                    loader.add(LONGITUDE, DEG, None, activity_group, ajournal,
                              record.none.position_long, record.value.timestamp,
                              StatisticJournalFloat)
                    if record.none.position_lat and record.none.position_long:
                        p = Point.from_latitude_longitude(record.none.position_lat, record.none.position_long)
                        x, y = p.meters
                        loader.add(SPHERICAL_MERCATOR_X, M, None, activity_group, ajournal,
                                  x, record.value.timestamp, StatisticJournalFloat)
                        loader.add(SPHERICAL_MERCATOR_Y, M, None, activity_group, ajournal,
                                  y, record.value.timestamp, StatisticJournalFloat)
                    loader.add(HEART_RATE, BPM, None, activity_group, ajournal,
                              record.none.heart_rate, record.value.timestamp, StatisticJournalInteger)
                    loader.add(DISTANCE, M, None, activity_group, ajournal,
                              record.none.distance, record.value.timestamp, StatisticJournalFloat)
                    loader.add(SPEED, KMH, None, activity_group, ajournal,
                              record.none.enhanced_speed, record.value.timestamp, StatisticJournalFloat)
                if record.name == 'event' and record.value.event == 'timer' \
                        and record.value.event_type == 'stop_all':
                    if timespan:
                        timespan.finish = record.value.timestamp
                        ajournal.finish = record.value.timestamp
                        timespan = None
                    else:
                        self._log.warn('Ignoring stop with no corresponding start (possible lost data?)')
                if record.name == 'record':
                    last_timestamp = record.timestamp
            else:
                if record.name == 'record':
                    self._log.warn('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                   (path, record.timestamp))

        loader.load()

        # manually clean out intervals because we're doing a stealth load
        Interval.clean_times(s, first_timestamp, last_timestamp)

        # used by subclasses
        return ajournal, loader





from .args import FORCE
from ..squeal.tables.constant import SystemConstant


def unlock(args, log, db):
    '''
## unlock

    > ch2 unlock --force

Remove any locking.

The database is locked to allow fast loading of data which requires no other command access the database.
Using this command removes the lock and so MAY CAUSE DATA CORRUPTION if the loading is still in progress.

You should only use this command in the unlikely case that somehow the lock remained after the loading finished
(eg. if the system crashed or was interrupted during loading).
    '''
    if not args[FORCE]:
        raise Exception('See `ch2 help unlock`')
    log.warn('Removing any database lock(s).  This may cause data corruption.  See `ch2 help unlock`')
    with db.session_context() as s:
        SystemConstant.reset_lock(s)
        log.info('Lock reset')




from math import pi, cos

from .tree import LinearMixin, BaseTree, QuadraticMixin, ExponentialMixin, CartesianMixin

RADIUS = 6371000
RADIAN = pi / 180


def norm180(x):
    while x > 180: x -= 360
    while x <= -180: x += 360
    return x


class LocalTangent:
    '''
    Assume a spherical earth and local linear approximations to convert from (lon, lat) to (x, y) in m.
    '''

    def __init__(self, point=None):
        self.__zero = None
        if point is not None:
            self.normalize(point)

    def normalize(self, point):
        if self.__zero is None:
            self.__zero = point
        zx, zy = self.__zero
        lon, lat = norm180(point[0] - zx), point[1] - zy
        return RADIUS * RADIAN * lon * cos(self.__zero[1]), RADIUS * RADIAN * lat

    def denormalize(self, point):
        zx, zy = self.__zero
        x, y = point
        return zx + x / (RADIUS * RADIAN * cos(self.__zero[1])), zy + y / (RADIUS * RADIAN)


class SphericalMixin(CartesianMixin):

    def __init__(self, *args, **kargs):
        self.__plane = LocalTangent()
        super().__init__(*args, **kargs)

    def _normalize_point(self, point):
        return self.__plane.normalize(point)

    def _denormalize_point(self, point):
        return self.__plane.denormalize(point)


class SLRTree(LinearMixin, SphericalMixin, BaseTree): pass


class SQRTree(QuadraticMixin, SphericalMixin, BaseTree): pass


class SERTree(ExponentialMixin, SphericalMixin, BaseTree): pass


# TODO - need to patch latitude too for tangent!!!

class GlobalLongitude:
    '''
    Cover longitude in overlapping local trees then query both the central tree and those to either side.

    We can either store three times and retrieve once, or store once and retrieve three times.
    '''

    def __init__(self, tree=SQRTree, n=36, favour_read=True):
        self.__tree = tree
        self.__n = n
        self.__trees = [None] * n
        self.__favour_read = favour_read

    def __delegate(self, i):
        if i >= self.__n: i -= self.__n
        if i < 0: i += self.__n
        if self.__trees[i] is None:
            tree = self.__tree()
            # set the local tangent to the centre point  TODO - check
            bin_width = 360 / self.__n
            lon = (i + 0.5) * bin_width - 180
            tree.add([(lon, 0)], None)
            tree.delete([(lon, 0)])
            self.__trees[i] = tree
        return self.__trees[i]

    def __delegates(self, points, read=True):
        lon = norm180(points[0][0])
        bin_width = 360 / self.__n
        i = int((lon + 180) // bin_width)
        yield self.__delegate(i)
        if self.__favour_read != read:
            yield self.__delegate(i-1)
            yield self.__delegate(i+1)

    def get(self, points, value=None, match=None, border=None):
        for delegate in self.__delegates(points):
            yield from delegate.get(points, value=value, match=match, border=border)

    def get_items(self, points, value=None, match=None, border=None):
        for delegate in self.__delegates(points):
            yield from delegate.get_items(points, value=value, match=match, border=border)

    def add(self, points, value, border=None):
        for delegate in self.__delegates(points, read=False):
            delegate.add(points, value, border=border)

    def add_all(self, items, border=None):
        for (points, value) in items:
            for delegate in self.__delegates(points, read=False):
                delegate.add(points, value, border=border)

    def delete(self, points, value=None, match=None, border=None):
        '''
        This purposefully does not return number deleted.  If you are relying on that value,
        your code should not be using this wrapper.
        '''
        for delegate in self.__delegates(points):
            delegate.delete(points, value=value, match=match, border=border)

    def delete_one(self, points, value=None, match=None, border=None):
        for delegate in self.__delegates(points):
            delegate.delete_one(points, value=value, match=match, border=border)

    def __getitem__(self, points):
        return self.get(points)

    def __setitem__(self, points, value):
        self.add(points, value)

    def __delitem__(self, points):
        self.delete(points)

    def __bool__(self):
        return any(bool(t) for t in self.__trees)




from os.path import join, dirname
from pickle import load, dump

import openpyxl as xls
from pkg_resources import resource_stream

from .messages import Messages
from .support import NullableLog
from .types import Types

PROFILE_NAME = 'global-profile.pkl'
PROFILE = []


def read_profile(log, path, warn=False):
    nlog = NullableLog(log)
    wb = xls.load_workbook(path)
    types = Types(nlog, wb['Types'], warn=warn)
    messages = Messages(nlog, wb['Messages'], types, warn=warn)
    return nlog, types, messages


def load_profile(log):
    if not PROFILE:
        log.debug('Unpickling profile')
        input = resource_stream(__name__, PROFILE_NAME)
        PROFILE.append(load(input))
        PROFILE[0][0].set_log(log)
    return PROFILE[0][1:]


def load_fit(log, fit_path, warn=False, profile_path=None):
    if profile_path:
        log.debug('Reading profile from %s' % profile_path)
        _nlog, types, messages = read_profile(log, profile_path, warn=warn)
    else:
        types, messages = load_profile(log)
    log.debug('Read profile')
    with open(fit_path, 'rb') as input:
        data =input.read()
    return data, types, messages


def pickle_profile(log, in_path, warn=False):
    log.info('Reading from %s' % in_path)
    nlog, types, messages = read_profile(log, in_path, warn=warn)
    out_path = join(dirname(__file__), PROFILE_NAME)
    nlog.set_log(None)
    log.info('Writing to %s' % out_path)
    with open(out_path, 'wb') as output:
        dump((nlog, types, messages), output)
    # test loading
    log.info('Test loading from %r' % PROFILE_NAME)
    log.info('Loaded %s, %s' % load_profile(log))
